#! /usr/local/bin/perl

use strict;
use warnings;
use POSIX qw/strftime/;

# Removes all personally identify-able information from an HL7 message.
# Don't trust this to really work as intended - It just tries its best.

# A global array of fields which need to be obscured.
use constant OBSCURE => qw/
	PID-2 PID-3 PID-5 PID-7 PID-9 PID-11 PID-13 PID-18 PID-19 PID-20
	NK1-2 NK1-4 NK1-5 NK1-6 NK1-12
	MRG-7
	IN1-16 IN1-19
	GT1-3 GT1-4 GT1-5 GT1-6 GT1-12 GT1-19
/;
# The HL7 organization's inability to count is compensated for in the
# handling of blanked MSH fields.  They are basically one-upped.

# Handle some initialization
my $file = $ARGV[0] || &print_usage("No argument supplied.");

exec("perldoc", $0) if ($file eq "-h");

&print_usage("Could not read $file : $!") if ((!-r $file) && ($file ne "-"));

# Build the regular expressions.
my $REs = [];

&build_REs($REs);

# Open the file and start parsing.
if ($file eq "-")
{
	open FILE, "<&STDIN" || &print_usage("Could not open STDIN : $!");
}
else
{
	open FILE, "<$file" || &print_usage("Could not open $file : $!");
}

# Process the file.
for my $record (<FILE>)
{
	my %blank;

	# Since we are looping over this it is worth studying.
	study $record;

	for (@$REs)
	{
		# Look for blocked fields.
		if ($record =~ /$_/)
		{
			# Add the entire field to the blanking array.
			# Add all components of the field to the blanking array.
			for (split /\^/, $2)
			{
				next unless length($_) > 1;
				$blank{$_} = &convert($_);
			}
		}
	}

	for (keys %blank)
	{
		# This clears out every instance of each string that
		# appears in a blocked segment.  This helps to remove
		# things like patient names from text fields.
		$record =~ s/($_)/$blank{$_}/ig;
	}

	print $record;
}

close FILE || die "Could not close $file : $!\n";

exit;

# Subroutines

# Convert fields to obscure into
sub convert
{
	my $s = shift;

	$s =~ tr/a-zA-Z/X/;
	$s =~ tr/0-9/8/;

	return $s;
}

# Print the program usage and exit.
sub print_usage
{
	my $message = shift;

	print "Error: $message\n\n" if $message;
	
print <<"EndOfText";
Usage:
	$0 -h
	$0 file.hl7 > out.hl7
	cat file | $0 - > out.hl7

See help (-h) for more details.
EndOfText

	exit;
}

# Compile the regular expressions from the array of fields.
sub build_REs
{
	my $RE_ref = shift;

	for (OBSCURE)
	{
		my ($seg, $field) = split /-/, $_;

		die "Can't blank a field-0\n" if $field == 0;

		$seg = uc $seg;

		if ($seg eq "MSH")
		{
			$field++;
			push @{$RE_ref}, '(^' . $seg . '(?:[^\|]*\|){' . $field . '})([^\|\r]+)';
		}
		else
		{
			push @{$RE_ref}, '(\r' . $seg . '(?:[^\|]*\|){' . $field . '})([^\|\r]+)';
		}
	}
}

=pod

=head1 NAME

hl7_anon - a program to remove identifying information from an HL7 message

=head1 SYNOPSIS

	hl7_anon in.hl7 > out.hl7
	cat file.hl7 | hl7_anon - > out.hl7
	hl7_anon in.hl7 | gview -
	cat logfile.txt | hl7_clean - | hl7_anon - > out.hl7

=head1 DESCRIPTION

hl7_anon reads in a file (or STDIN) with HL7 records and attempts to remove
all identifying information from the record.  It does so by looking for HL7
fields specified in it's list and blanking those fields out.  It also adds
the contents of those fields (and the components of those fields where
applicable) to an array which will then be used to find instances of that
private information anywhere in the record and blank it there.

It is important to realize that this program is not a perfect means of
removing personal information from an HL7 document and manual checking should
always be done of the information that might be seen by prying eyes.

How this software works is that it scans the entire record looking for fields
that should be hidden.  Upon finding one it is split into components (or
taken as a single record of no components are present).  Components that are
shorter than two characters in length are discarded.  The components are
scanned and each alpha character is converted into the letter "X" and each
numeric character is converted into the number "8".  The converted strings
are stored in a data structure along with the original string and a global
search and replace is performed on the record.

Because the matching and replacing occur in different steps, the end effect
is that any information that should be modified in the record is modified
everywhere it occurs in the record (even outside of the specified fields),
as long as the spelling is the same (case is ignored on the final replace).
This is handy for things like transcription reports where the patient name
and MRN are also in the OBX fields as part of the report.

=head1 ARGUMENTS

The only argument you need to give hl7_anon is the name of the file you wish
it to process.  If the file-name is a dash it will read the records from
standard in.  This is good if you wish to pipe things through hl7_clean or
from a bzipped file first.  In particular hl7_clean might come in handy here
as this program has no logic to read HL7 information from within a log-file
as hl7_clean does.

hl7_anon also takes a '-h' as an option to print this helpful information
you are now viewing.

=head1 AUTHOR

M. Litherland (litherm@ccf.org)

=head1 HISTORY

=over 4

=item 2003-03-14

Initial implementation (ml)

=back

=cut

